#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import glob
import logging
import fnmatch

import lbuild.utils

from .facade import RepositoryInitFacade, RepositoryPrepareFacade
from .node import BaseNode
from .exception import LbuildException

LOGGER = logging.getLogger('lbuild.repository')


class Repository(BaseNode):
    """
    A repository is a set of modules.
    """

    def __init__(self, filename, name=None):
        """
        Construct a new repository object.

        At the construction time of the object, the name of repository may not
        be known e.g. if the repository is loaded from a `repo.lb` file.
        """
        BaseNode.__init__(self, name, self.Type.REPOSITORY, self)
        # Path to the repository file. All relative paths refer to this path.
        self._filename = os.path.realpath(filename)
        self._config_map = {}
        self._new_filters = {}

        # List of module filenames which are later transfered into
        # module objects
        self._module_files = []

    @property
    def modules(self):
        return {m.fullname:m for m in self.all_modules()}

    @staticmethod
    def parse_repository(repofilename: str):
        LOGGER.debug("Parse repository '%s'", repofilename)

        repo = Repository(repofilename)
        repo._functions = lbuild.node.load_functions_from_file(
            repo, repofilename, required=['init', 'prepare'], optional=['build'])

        # Execution init() function. In this function options are added.
        repo._functions['init'](RepositoryInitFacade(repo))

        if repo.name is None:
            raise LbuildException("The init(repo) function must set a repository name! "
                                  "Please write the 'name' attribute.")

        # Prefix the global filters with the `repo.` name
        for name, func in repo._new_filters.items():
            repo._filters[repo.name + "." + name] = func

        # Prefix the global configuration map with the `repo:` name
        config_map = repo._config_map.copy()
        repo._config_map = {}
        for name, path in config_map.items():
            repo._config_map[repo.name + ":" + name] = path

        return repo

    def prepare(self):
        lbuild.utils.with_forward_exception(
            self,
            lambda: self._functions["prepare"](RepositoryPrepareFacade(self),
                                               self.option_value_resolver))

        modules = []
        # Parse the modules inside this repository
        for modulefile in self._module_files:
            module = lbuild.module.load_module_from_file(self, modulefile)
            modules.extend(module)

        return modules

    def build(self, env):
        build = self._functions.get("build", None)
        if build is not None:
            lbuild.utils.with_forward_exception(self, lambda: build(env))

    def add_modules_recursive(self, basepath="", modulefile="module.lb", ignore=None):
        """
        Find all module files following a specific pattern.

        Args:
            basepath: Rootpath for the search.
            modulefile: Filename pattern of the module files to search
                for (default: "module.lb").
            ignore: Filename pattern to ignore during search
        """
        ignore = lbuild.utils.listify(ignore) + self._ignore_patterns
        ignore.append(os.path.relpath(self._filename, self._filepath))
        basepath = self._relocate_relative_path(basepath)
        for path, _, files in os.walk(basepath):
            for file in files:
                if any(fnmatch.fnmatch(file, i) for i in ignore):
                    continue
                if fnmatch.fnmatch(file, modulefile):
                    modulefilepath = os.path.normpath(os.path.join(path, file))
                    self._module_files.append(modulefilepath)

    def add_modules(self, *modules):
        """
        Add one or more module files.

        Args:
            modules: List of filenames
        """
        modules = [lbuild.utils.listify(m) for m in modules]
        modules = [inner for outer in modules for inner in outer]
        for file in modules:
            file = self._relocate_relative_path(file)

            if not os.path.isfile(file):
                raise LbuildException("Module file not found '%s'" % file)

            self._module_files.append(file)

    def glob(self, pattern):
        pattern = os.path.abspath(self._relocate_relative_path(pattern))
        return glob.glob(pattern)

    def __lt__(self, other):
        return self.fullname.__lt__(other.fullname)

    def __repr__(self):
        return "Repository({})".format(self._filepath)

    def __str__(self):
        """ Get string representation of repository.  """
        if self.name is not None:
            # The name must not always be set, e.g. during load
            # the repository name is only known after calling the
            # init function.
            return "{} at {}".format(self.name, self._filepath)

        return self._filepath



#!/usr/bin/env python3
#
# Copyright (c) 2015, 2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

from lbuild.main import __version__

from . import builder
from . import buildlog
from . import environment
from . import exception
from . import module
from . import option
from . import parser
from . import repository
from . import utils
from . import main

__all__ = [
    'builder',
    'buildlog',
    'environment',
    'exception',
    'facade',
    'module',
    'node',
    'option',
    'parser',
    'repository',
    'utils',
    'main'
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2018, Niklas Hauser
# Copyright (c) 2018, Fabian Greif
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

"""
API for scripts extending/querying lbuild directly.
"""

import os

import lbuild.environment
from lbuild.buildlog import BuildLog
from lbuild.parser import Parser
from lbuild.config import ConfigNode
from lbuild.utils import listify, listrify


class Builder:

    def __init__(self, cwd=None, config=None, options=None):
        """
        Build instance to invoke lbuild methods.

        Args:
            cwd -- Current working directly. If specified lbuild will
                search for a configuration file in this folder.
            config -- Path to a configuration file. If specified options,
                repositories etc. will be loaded from there.
            options -- List of options. Must be list of strings in a
                "key=value" format.
        """
        if cwd is None:
            cwd = os.getcwd() if config is None else os.path.dirname(config)
        self.cwd = cwd

        try:
            config = ConfigNode.from_file(config)
        except lbuild.exception.LbuildException:
            config = None

        file_config = ConfigNode.from_filesystem(cwd)
        if config:
            config.extend_last(file_config)
        else:
            config = file_config if file_config else ConfigNode()

        self.config = config
        self.config.add_commandline_options(listify(options))
        self.parser = Parser(self.config)

    def _load_repositories(self, repos=None):
        self.parser.load_repositories(listrify(repos))
        self.parser.merge_repository_options()

    def _load_modules(self):
        if not self.parser._undefined_repo_options():
            self.parser.prepare_repositories()
            self.parser.merge_module_options()

    def _filter_modules(self, modules=None):
        self.parser.config.modules.extend(listify(modules))
        selected_modules = self.parser.find_modules(self.parser.config.modules)
        return self.parser.resolve_dependencies(selected_modules)

    def load(self, repos=None):
        """
        Load a list of repositories.

        Args:
            repos -- List of paths to repostiory files. Modules found
                in these repositores will be available later.
        """
        self._load_repositories(repos)
        self._load_modules()

    def validate(self, modules=None):
        """
        Generate and validate the required set of modules.

        Checks that the modules could be built, but does not generate
        the output.

        Args:
            modules -- List of modules which should be validated.

        Returns:
            List of modules required for the given modules (after
            resolving dependencies).
        """
        build_modules = self._filter_modules(modules)
        self.parser.validate_modules(build_modules)
        return build_modules

    def build(self, outpath, modules=None, simulate=False):
        """
        Build the given set of modules.

        Args:
            outpath -- Path where the output will be generated.
            modules -- List of modules which should be built. This list
                is combined with modules given in the configuration
                files.
            simulate -- If set to True simulate the build process. In
                that case no output will be generated.
        """
        build_modules = self._filter_modules(modules)
        buildlog = BuildLog(outpath)
        lbuild.environment.SIMULATE = simulate
        self.parser.build_modules(build_modules, buildlog)
        return buildlog



#!/usr/bin/env python3
#
# Copyright (c) 2018, Niklas Hauser
# Copyright (c) 2018, Fabian Greif
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import lbuild.utils


class BaseNodePrepareFacade:

    def __init__(self, node):
        self._node = node

    @property
    def name(self):
        return self._node.name

    @property
    def description(self):
        return self._node.description

    @property
    def format_description(self):
        return self._node._format_description

    @property
    def format_short_description(self):
        return self._node._format_short_description


class BaseNodeInitFacade(BaseNodePrepareFacade):

    def __init__(self, node):
        BaseNodePrepareFacade.__init__(self, node)

    # Disable warnings caused by property setters which are not properly recognised by pylint
    # pylint: disable=no-member
    @BaseNodePrepareFacade.name.setter
    def name(self, value):
        self._node.name = value
        self._node._fullname = value

    @BaseNodePrepareFacade.description.setter
    def description(self, value):
        self._node.description = value

    @BaseNodePrepareFacade.format_description.setter
    def format_description(self, formatter):
        self._node._format_description = formatter

    @BaseNodePrepareFacade.format_short_description.setter
    def format_short_description(self, formatter):
        self._node._format_short_description = formatter


class RepositoryInitFacade(BaseNodeInitFacade):

    def __init__(self, repository):
        BaseNodeInitFacade.__init__(self, repository)

    def add_option(self, option):
        self._node.add_option(option)

    def add_ignore_patterns(self, *patterns):
        self._node._ignore_patterns.extend(patterns)

    def add_filter(self, name, function):
        self._node._new_filters[name] = function

    def add_configuration(self, name, path):
        self._node._config_map[name] = path


class RepositoryPrepareFacade(BaseNodePrepareFacade):

    def __init__(self, repository):
        BaseNodePrepareFacade.__init__(self, repository)

    def add_modules_recursive(self, basepath="", modulefile="module.lb", ignore=None):
        self._node.add_modules_recursive(basepath, modulefile, ignore)

    def add_modules(self, *modules):
        self._node.add_modules(*modules)

    def glob(self, pattern):
        return self._node.glob(pattern)

    # deprecated
    def find_modules_recursive(self, basepath="", modulefile="module.lb", ignore=None):
        self.add_modules_recursive(basepath, modulefile, ignore)


class ModulePrepareFacade(BaseNodePrepareFacade):

    def __init__(self, module):
        BaseNodePrepareFacade.__init__(self, module)

    @property
    def parent(self):
        return self._node.parent

    def add_option(self, option):
        self._node._options.append(option)

    def add_submodule(self, module):
        self._node._submodules.append(module)

    def depends(self, *dependencies):
        self._node._dependencies.extend(dependencies)


class ModuleInitFacade(BaseNodeInitFacade):

    def __init__(self, module):
        BaseNodeInitFacade.__init__(self, module)

    @property
    def parent(self):
        return self._node.parent

    @parent.setter
    def parent(self, name):
        self._node.parent = name

    def add_filter(self, name, function):
        self._node._filters[name] = function


class EnvironmentValidateFacade:

    def __init__(self, env):
        self._env = env

    @property
    def log(self):
        return self._env.log

    def has_option(self, key):
        return self._env.has_option(key)

    def has_module(self, key):
        return self._env.has_module(key)

    def get(self, key, default=None):
        return self._env.get_option(key, default)

    def __getitem__(self, key):
        return self._env.__getitem__(key)

    # deprecated
    def get_option(self, key, default=None):
        return self.get(key, default)


class EnvironmentPostBuildFacade(EnvironmentValidateFacade):

    def __init__(self, env):
        EnvironmentValidateFacade.__init__(self, env)

    @property
    def outbasepath(self):
        return self._env.outbasepath

    @outbasepath.setter
    def outbasepath(self, path):
        self._env.outbasepath = path

    @property
    def substitutions(self):
        return self._env.substitutions

    @substitutions.setter
    def substitutions(self, substitutions):
        self._env.substitutions = substitutions

    def copy(self, src, dest=None, ignore=None):
        self._env.copy(src, dest, ignore)

    def extract(self, archive, src=None, dest=None, ignore=None):
        self._env.extract(archive, src, dest, ignore)

    def template(self, src, dest=None, substitutions=None, filters=None):
        self._env.template(src, dest, substitutions, filters)

    def generated_local_files(self, filterfunc=None):
        return self._env.generated_local_files(filterfunc)

    def reloutpath(self, path, relative=None):
        return self._env.reloutpath(path, relative)

    @staticmethod
    def ignore_files(*files):
        return lbuild.utils.ignore_files(*files)

    @staticmethod
    def ignore_paths(*paths):
        return lbuild.utils.ignore_patterns(*paths)

    # deprecated
    @staticmethod
    def ignore_patterns(*patterns):
        return lbuild.utils.ignore_patterns(*patterns)

    # deprecated
    def get_generated_local_files(self, filterfunc=None):
        return self.generated_local_files(filterfunc)

    # deprecated
    def outpath(self, *path, basepath=None):
        return self._env.outpath(*path, basepath=basepath)


class EnvironmentBuildFacade(EnvironmentPostBuildFacade):

    def __init__(self, env):
        EnvironmentPostBuildFacade.__init__(self, env)

    def add_metadata(self, key, *values):
        self._env.add_metadata(key, *values)

    # deprecated
    def append_metadata(self, key, *values):
        self._env.add_metadata(key, *values)

    # deprecated
    def append_metadata_unique(self, key, *values):
        self._env.add_metadata(key, *values)


class BuildLogOperationFacade:

    def __init__(self, operation):
        self._operation = operation

    @property
    def module_name(self):
        return self._operation.module_name

    def filename_out(self, path=None):
        return self._operation.local_filename_out(path)


class BuildLogFacade:

    def __init__(self, buildlog):
        self._buildlog = buildlog

        self.__metadata = buildlog.metadata
        self.__repo_metadata = buildlog.repo_metadata
        self.__module_metadata = buildlog.module_metadata

    @property
    def outpath(self):
        return self._buildlog._outpath

    @property
    def metadata(self):
        return self.__metadata

    @property
    def module_metadata(self):
        return self.__module_metadata

    @property
    def repo_metadata(self):
        return self.__repo_metadata

    @property
    def modules(self):
        return self._buildlog.modules

    @property
    def operations(self):
        return self._buildlog.operations

    def operations_per_module(self, modulename: str):
        return self._buildlog.operations_per_module(modulename)

    def __iter__(self):
        return iter(self._buildlog.operations)

    # deprecated
    def get_operations_per_module(self, modulename: str):
        return self.operations_per_module(modulename)
